#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2014-2015 Japan Atomic Energy Agency
#
#    angel2vtk is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation;
#    version 2.1 of the License.
#
#    This software is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this software; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Author:
#       Kazuyoshi Furutaka <furutaka.kazuyoshi@jaea.go.jp>

"""The main module in angel2vtk papckage.

This module contains classes for decoding ANGEL_ input files.

* :py:class:`angelFile`
* :py:class:`PHITS_TTrackData`
* :py:class:`convertThread`
* :py:class:`logBuffer`
* :py:class:`Angel2VTK_MainWindow`
* :py:class:`LogStream`

This module also contains main functions for PyQt4_ GUI together with the
one for CUI which are used as entry points for scripts created by
SETUPTOOLS_ :

* :py:func:`GUImain`
* :py:func:`main`

This module, if run as a standalone program, converts each input
'Angel' file named on the command line to (intermediate ROOT_ file
and then) a group of VTK_ image files (.vti) and the summary file
(.pvd).

In addition to PyQt4_ mentioned above, this module depends on the
following external program and packages:

* ROOT_
* PyEVTK_
* mc-tools_

Though efforts have been paid to adopt to Py3k_ it is not tested
and may well be imperfect.

To do (almost) real-time logging, the author of this package
have tried to use a few methods:

* :py:class:`logBuffer`
* :py:class:`LogStream`

but he finally decided to use Python's :py:mod:`logging` facility
with :py:mod:`threading` (that is, the two classes for logging are
*not used*);
the important thing is passing of (logging) messages in
separate threads!  Otherwise, the messages would be passed
after finishing the corresponding procedures.

.. _ANGEL:      http://phits.jaea.go.jp/manual/manualJ-angel4.pdf
.. _PyQt4:      http://www.riverbankcomputing.co.uk/software/pyqt/intro
.. _ROOT:       http://root.cern.ch
.. _VTK:        http://www.vtk.org
.. _SETUPTOOLS: https://pypi.python.org/pypi/setuptools
.. _PyEVTK:     https://bitbucket.org/pauloh/pyevtk
.. _mc-tools:   https://github.com/kbat/mc-tools
.. _Py3k:       http://legacy.python.org/dev/peps/pep-3000/
"""

import re, sys, os, math, platform
import argparse
import logging, threading, time
#: A boolean which describes whether it is run in Python 3.x.
Py3k = True if sys.version_info.major >= 3 else False
# import io as IO if Py3k else import StringIO as IO # cause error in 2to3
if Py3k:
    import io as IO #if Py3k
else:
    import StringIO as IO

import numpy as np
from PyQt4.QtCore import *
from PyQt4.QtGui import *
import ROOT
# The line below is needed to prevent command-line arguments from
# stolen by PyROOT and handed to TApplication
ROOT.PyConfig.IgnoreCommandLineOptions = True
from evtk.hl import imageToVTK
from evtk.vtk import VtkGroup

import ui_angel2vtk
#import qrc_resources

__version__ = "0.0.1"

MAC = "qt_mac_set_native_menubar" in dir()

#: a regular expression which describes the page-separating line.
pageSepRE = re.compile("^[\s#]newpage:$")

#: a regular expression which describes the section-separating line on 0th page.
page0SecSepRE = re.compile("^(([xyzet])-type|unit) =\s+(\d{1,2})\s+#\s+(.+)")
# match object number        ((   1   )   0      )     ( 2     )       ( 3)
#=>groups()=('x-type', 'x', '2', 'x-mesh is linear given by xmin, xmax and nx')
#           ('unit', None, '1', 'unit is [1/cm^2/source]')

#: a regular expression which describes the value line?
valueRE = re.compile("^#?\s*[a-z]+ =\s+(.*)\s+#\s+.*")  # group( 1 ) only
#: a regular expression which describes the type of the mesh
typeRE =  re.compile("^[xyzet]-type =\s+(.*)\s+#\s+.*") # group( 1 ) only

DEBUG = False
#DEBUG = True

class angelFile:
    """Read an angle file named ``inFileName`` and store the contents.

    attributes are:
    
    * :py:data:`pageLST`     (private)
    * :py:data:`numNewpages` (private)
    * :py:meth:`getNumPages` (public)
    * :py:meth:`getPage`     (public)
    """
    #: the name of the input file
    inFileName = None
    #: a python list which contains the numbers of lines which separate
    #: pages
    pageSepLineLST = []
    #: a list of tuples; each tuple contains the contents of a page.
    pageLST = []

    def __init__( self, inFileName ):
        """

        0. create a logger,
        1. open, read all the contents of the named file and close it,
        2. search for page separators and store their offsets,
        3. store the contents of each page as a tuple in a list named ``pageLST`.
        """

        #: the number of ``newpage:``.
        numNewpages = 0

        # logging
        self.logger = logging.getLogger(__name__)
        self.logger.setLevel(logging.INFO)
        #self.logger.setLevel(logging.DEBUG)

        # open the named input file, read all the lines in the file at once,
        # and process them
        inFile = open(inFileName,'r') # defaults to 'r' but explicitly...
        lines = inFile.readlines()
        inFile.close()

        # scan and count the "newpage:", 
        for idx,line in enumerate(lines):
            modLine = line.rstrip() # strip spaces/newlines at the right
            if pageSepRE.search(modLine):
                numNewpages += 1
                self.pageSepLineLST.append(idx)
        self.logger.debug("number of '[#]newpage:' = %d", numNewpages)
        self.logger.debug("number of pages including the preamble(info) = %d",
                          numNewpages+1)
        self.logger.debug("pageSepLineLST = %d", self.pageSepLineLST)

        # Separate each page into a tuple and
        # store them into the pageLST
        #   the 1st page
        self.pageLST.append( tuple( lines[:self.pageSepLineLST[0]] ) )
        #   intermediate
        for pageIdx in range(1,numNewpages):
            self.pageLST.append(
                lines[ self.pageSepLineLST[pageIdx-1]+1 :
                       self.pageSepLineLST[pageIdx] ] )
        #   the last page
        self.pageLST.append( tuple( lines[self.pageSepLineLST[-1]+1:] ) )
        del lines # delete needless storage
        for lineIdx,line in enumerate(self.pageLST[0]):
            self.logger.debug("%d %s",lineIdx, line.rstrip())
        
    def getNumPages(self):
        """Return the number of pages in the ANGEL_ file read.

        (=the number of '#?newpage:' +1)
        """
        return len( self.pageLST ) # should be == self.numNewpages+1


    def getPage(self,page):
        """Returns a list that contains the information in the named ``page``.
        """
        if page < 0 or page >= len( self.pageLST ):
            return None
        else:
            return self.pageLST[ page ]

class PHITS_TTrackData:
    """A class for decoding the info in a PHITS T-Track output file (i.e. an input for 'Angel').

    :Parameter:
    :py:data:`AngelFile`
      the path to an input angel file to be converted.
    """
    #: jobTitle: self-evident
    jobTitle = None
    #: string which describes the unit of the mesh values, e.g. "[1/cm^2/source]".
    unit = None # e.g. "[1/cm^2/source]"
    #: axes along which the mesh values are scored, e.g.``xy``, ``yz``, ``xz``,
    #: or ``rz``.
    axis = None # one of  "xy", "yz", "xz", "rz" ???
    #: a dictionary which store the mesh information decoded in :py:meth:`decodeMeshInfo`.
    mesh = {}

    def __init__(self, AngelFile):
        """The initializer of the class.

        0. create a :py:mod:`logging` logger,
        1. look at the first page of the ``Angel`` data,
        2. make sure it's a PHITS_ ``T-Track`` ``xyz-mesh`` tally, 
        3. find the starting line of each axis and store their offsets,
        """
        # logging
        self.logger = logging.getLogger(__name__)
        self.logger.setLevel(logging.INFO)
        #self.logger.setLevel(logging.DEBUG)

        firstPageLines = AngelFile.getPage(0)
        # check that it's a T-Track data
        if firstPageLines[0].strip() != "[ T - T r a c k ]":
            self.logger.warning("It's NOT a T-Track TALLY, exitting...")
            return None
        # get the job title
        lineStr = firstPageLines[1].strip()
        if lineStr[0:5] == "title":
            self.jobTitle = lineStr[8:]
            self.logger.debug("%s", self.jobTitle)
        else:
            self.logger.warning("the 2nd line NOT 'title', exitting...")
            return None
        # check that it's a xyz-mesh data
        lineStr = firstPageLines[2].strip()
        if lineStr[0:4] != "mesh":
            self.logger.warning("the 3rd line NOT 'mesh', exitting...")
            return None
        elif lineStr[8:11] != "xyz":
            self.logger.warning("CAN NOT process mesh other than 'xyz', exitting...")
            sys.exit(1)
        # extract [xyzet] mesh-grid, unit, and axis information.
        #   first, extract the numbers of dividing lines
        #   ( "[xyzet]-type = ..." or "unit = "
        # the order is assumed to be x -> y -> z -> e -> t -> unit
        axisLineOffset = []
        axisOrder  = ['x', 'y', 'z', 'e']
        # search the section-separating lines
        for idx,line in enumerate(firstPageLines):
            modLine = line.strip() # strip spaces/newlines
            mObj = page0SecSepRE.search(modLine) # match object of section-separator
            if mObj:
                moTup = mObj.groups()
                self.logger.debug("%d %s", idx, modLine)
                axisLineOffset.append( idx )
                if not moTup[1]: # moTup[1] != None, i.e. 'unit' line
                    self.unit = moTup[3][8:]
                    self.axis = firstPageLines[idx+3].strip()[9:11]
                elif moTup[1] == 't':
                    axisOrder.append('t')
        self.logger.debug("axisLineOffset = %d", axisLineOffset)
        self.logger.debug("unit = '%s', axis = '%s'", self.unit, self.axis)
        # decode information in each mesh/section
        for idx,ax in enumerate(axisOrder):
            # print "ax, type(ax) = ", ax, type(ax)
            self.decodeMeshInfo( ax, firstPageLines[ axisLineOffset[idx] :
                                                     axisLineOffset[idx+1] ] )
        self.logger.debug("axis = %s", self.axis)
        for Axis in [i for i in self.mesh.keys()]: # Py3k
            self.logger.debug("%s : %s", Axis, self.mesh[ Axis ])

    def decodeMeshInfo(self, Axis, lineTup):
        """Decode the info for the named ``Axis`` in the tuples (``lineTup``) passed, and store them in a dictionary.

        This is a helper method invoked from the ``__init__``.

        The keys and the corresponding values in the resultant dictionary may be:

        ``type``
            a number describing the type of the mesh (implemented only for ``1`` and ``2``)

            ``1``
               specify the number of groups and division points
            ``2``
               specify the number of groups as well as minimum and maximum (linear)
            ``3``
               specify the number of groups as well as minimum and maximum (log)
            ``4``
               specify width of meshes and minimum, maximum (linear)
            ``5``
               specify (log-)width of meshes and minimum, maximum (log)

        ``num`` (the corresponding value is an integer), 
          the number of bins (mesh divisions)
        ``min``
          the lower edge of the lowest bin
        ``max``
          the upper edge of the highest bin
        ``sep``
          a float which describes the width/separation of the bins
        ``edge``
          a list of floats which describe bounds of the bins
        """
        # meshType = 1-5
        meshType = int( typeRE.search( lineTup[0].strip() ).group( 1 ) )
        if meshType == 1: # number, edge data
            numVal =  int(   valueRE.search( lineTup[1].strip() ).group( 1 ) )
            edges = tuple([float(val) for val in
                           ''.join(lineTup[3:]).replace('#','').replace('\r','')
                           .replace('\n','').split() ])
            self.mesh[ Axis ] = {'type':meshType,
                                 'min':edges[0], 'max':edges[-1],
                                 'num':numVal, 'edge':edges}
        elif meshType == 2: # number, min, max (linear)
            minVal =  float( valueRE.search( lineTup[1].strip() ).group( 1 ) )
            maxVal =  float( valueRE.search( lineTup[2].strip() ).group( 1 ) )
            delVal =  float( valueRE.search( lineTup[3].strip() ).group( 1 ) )
            numVal =  int(   valueRE.search( lineTup[4].strip() ).group( 1 ) )
            edges = tuple([float(val) for val in 
                           ''.join(lineTup[6:]).replace('#','').replace('\r','')
                           .replace('\n','').split() ])
            self.mesh[ Axis ] = {'type':meshType,
                                 'min':minVal, 'max':maxVal,
                                 'sep':delVal,
                                 'num':numVal, 'edge':edges}
        else:
            self.logger.warning("meshType = %d NOT YET IMPLEMENTED", meshType)

    def getJobTitle(self):
        """Returns a string which contains the title of the job."""
        return self.jobTitle

    def getUnitInfo(self):
        """Returns a string which describes the unit of the tally."""
        return self.unit

    def getMeshInfo(self, Axis):
        """Returns a dictionary that contains information on the named ``Axis``."""
        return self.mesh[ Axis ]

    def getAxisInfo(self):
        """Returns the information on the mesh axes.

        (e.g. "eng", "reg", "x", "y", "z", "r", "xy", "yz", "xz", "rz", "t").
        """
        return self.axis

    def getNumBins(self,Axis):
        """Returns the number of bins in the named ``Axis``.

        The ``Axis`` should be one of ``x``, ``y``, ``z``, ``e``, and ``t``.
        """
        if Axis in [i for i in self.mesh.keys()]: # Py3k
            return self.mesh[Axis]['num']
        else:
            return None

    def getNumBins(self):
        """Returns a tuple that contains the numbers of bins for all axis.

        ``Axis`` = (``x``, ``y``, ``z``, ``e``, and ``t`` if it exists)
        """
        if 't' in self.mesh:
            return (self.mesh['x']['num'], self.mesh['y']['num'], 
                    self.mesh['z']['num'], self.mesh['e']['num'], 
                    self.mesh['t']['num'] )
        else:
            return (self.mesh['x']['num'], self.mesh['y']['num'], 
                    self.mesh['z']['num'], self.mesh['e']['num'], 1 )

    def getBinWidth(self,Axis):
        """Returns the value of the (common) bin width for the named ``Axis`` if it makes sense."""
        Type = self.mesh[Axis]['type']
        Num  = self.mesh[Axis]['num']
        if Type == 1 and  Num == 1:
            return self.mesh[Axis]['max'] - self.mesh[Axis]['min']
        elif Type == 2:
            return self.mesh[Axis]['sep']
        else:
            self.logger.warning("bin width meaningless for "
                                "(axis,type,num)=(%s,%d,%d)", Axis,Type,Num)
            return None

    def getMinXYZ(self):
        """Returns a tuple that contains the minimam values for axes ``x``, ``y``, and ``z``."""
        return (self.mesh['x']['min'], self.mesh['y']['min'], 
                self.mesh['z']['min'])
        
    def getTypesXYZ(self):
        """Returns a tuple that contains the types for axes ``x``, ``y``, and ``z``."""
        return (self.mesh['x']['type'], self.mesh['y']['type'], 
                self.mesh['z']['type'])

    def getTypesXYZT(self):
        """Returns a tuple that contains the types for axes ``x``, ``y``, ``z``, and ``t``."""
        return (self.mesh['x']['type'], self.mesh['y']['type'], 
                self.mesh['z']['type'], self.mesh['t']['type'])

    def getTypes(self):
        """Returns a tuple that contains the types for all axes (``x``, ``y``, ``z``, ``e``, ``t``)."""
        if 't' in self.mesh:
            return (self.mesh['x']['type'], self.mesh['y']['type'], 
                    self.mesh['z']['type'], self.mesh['e']['type'],
                    self.mesh['t']['type'])
        else:
            return (self.mesh['x']['type'], self.mesh['y']['type'], 
                    self.mesh['z']['type'], self.mesh['e']['type'], None )

    # edge 0   1   2   3   4   5   6   7   8   9  10   num num+1
    # bin  | 0 | 1 | 2 | 3 | 4 | 5 | 6 | 7 | 8 | 9 | ...|num|
    #      +---+---+---+---+---+---+---+---+---+---+ ...+---+
    def getUEdge(self,Axis,bin):
        """Return the upper bound of the named ``bin`` of the named ``Axis``."""
        if bin >= 0 and bin < self.mesh[Axis]['num']:
            return self.mesh[Axis]['edge'][bin+1]
        else:
            return None
        
    def getLEdge(self,Axis,bin):
        """Return the lower bound of the named ``bin`` of the named ``Axis``."""
        if bin >= 0 and bin < self.mesh[Axis]['num']:
            return self.mesh[Axis]['edge'][bin]
        else:
            return None

class Angel2VTK_MainWindow(QMainWindow, ui_angel2vtk.Ui_MainWindow):
    """To instantiate and manipulate the MainWindow created by using `Qt Designer`_.

    :Parameters:

    :py:data:`filelist`
      a list of paths of files to be converted.
    :py:data:`digF`
      a Boolean specifying whether to dig a folder for each file or not.
    :py:data:`remR`
      a Boolean specifying whether to remove the intermediate ROOT_ files.

    .. _`Qt Designer`: http://qt-project.org/doc/qt-4.8/designer-manual.html
    """

    __digFolders = False
    __removeRoots = False
    __running = False
    __filename = None

    def __init__(self, parent=None, filelist=[], digF=False, remR=False):
        """Initializer of the class.

        Setup a logger, connect() some signals to slots, restore the previous settings,
        and prepare the list of files named on the command line.
        """
        super(Angel2VTK_MainWindow, self).__init__(parent)
        self.setupUi(self)
        if not MAC:
            pass

        # logging using a logBuffer instance
        # A part of the log capture method described in
        # http://stackoverflow.com/questions/14349563/how-to-get-non-blocking-real-time-behavior-from-python-logging-module-output-t
        self.logBuffer = logBuffer()
        self.logBuffer.bufferMessage.connect(self.on_logBuffer_bufferMessage)
        logFormatter = logging.Formatter("%(message)s")
        logHandler = logging.StreamHandler(self.logBuffer)
        logHandler.setFormatter(logFormatter)
        self.logger = logging.getLogger(__name__)
        self.logger.setLevel(logging.INFO)
        #self.logger.setLevel(logging.DEBUG)
        self.logger.addHandler(logHandler)
    
        # A part of the log capture method described in
        # http://stackoverflow.com/questions/8356336/how-to-capture-output-of-pythons-interpreter-and-sho-in-a-text-widget
        # # redirect sys.std{out,err}
        # sys.stdout = LogStream(textWritten=self.writeLogOutput)
        # sys.stderr = LogStream(textWritten=self.writeLogOutput)

        # # help menu on the rightmost corner
        # self.menubar.setStyle(QMotifStyle())

        # context menu for the file list
        self.inputFileList.addAction(self.actionOpen)

        # connect Quit -> close event
        self.connect(self.actionQuit, SIGNAL("triggered()"), self.close)

        ### Restore the application state using QSettings
        settings = QSettings()
        # size of the window
        size = settings.value("MainWindow/Size",
                              QVariant(QSize(600,500))).toSize()
        self.resize(size)
        # position of the window
        position = settings.value("MainWindow/Position",
                                      QVariant(QPoint(0,0))).toPoint()
        self.move(position)
        # digFolders & removeRoots "checked" or not
        # command-line options override the previous settings
        self.logger.debug("digF,remR='%s','%s'",digF,remR)
        self.__digFolders = settings.value("DigFolders",False).toBool() or digF
        self.actionDigFolders.setChecked(self.__digFolders)
        self.digFolderCheckBox.setChecked(self.__digFolders)
        self.__removeRoots = settings.value("RemoveRoots",False).toBool() or remR
        self.actionRemoveRoots.setChecked(self.__removeRoots)
        self.removeRootCheckBox.setChecked(self.__removeRoots)

        # dropped(?) files
        self.logger.debug("Angel2VTK_MainWindow.__init__: filelist = %s", filelist)
        for inFile in filelist:
            if not self.inputFileList.findItems(inFile,Qt.MatchExactly):
                self.inputFileList.addItem(inFile)
        
    @pyqtSlot(str)
    def on_logBuffer_bufferMessage(self, message):
        """capture the messages emitted from logBuffer instances.

        This is a part of the logging method described in `an article`_ .

        .. _`an article`: http://stackoverflow.com/questions/14349563/how-to-get-non-blocking-real-time-behavior-from-python-logging-module-output-t
        """
        self.logTextEdit.appendPlainText(message)

    # def writeLogOutput(self, text):
    #     """
    #
    #     This is a part of the logging method described in
    #     http://stackoverflow.com/questions/8356336/how-to-capture-output-of-pythons-interpreter-and-show-in-a-text-widget
    #     """
    #     self.logTextEdit.appendPlainText(text)
    #     #self.logTextEdit.appendPlainText(text.rstrip())
    #     self.logTextEdit.update()

    # def __del__(self):
    #     """
    #
    #     This is a part of the logging method described in
    #     http://stackoverflow.com/questions/8356336/how-to-capture-output-of-pythons-interpreter-and-show-in-a-text-widget
    #     """
    #     # STANDARD std{out,err}
    #     sys.stdout = sys.__stdout__
    #     sys.stderr = sys.__stderr__

    @pyqtSlot(bool)
    def on_actionOpen_triggered(self):
        """Respond to the trrigered signal of the ``Angel2VTK_MainWindow.actionOpen`` menu item.

        Just call the :py:meth:`fileOpen` method.
        """
        self.fileOpen()

    # def on_inputFileList_contextMenuEvent(self,event):
    #     print "contextMenuEvent"

    def fileOpen(self):
        """Select files to convert using QFileDialog.getOpenFileNames().

        Only the files w/ "*.out" extension can be selected.
        The results are appended to the ``Angel2VTK_MainWindow.inputFileList``.
        """
        dir = os.path.dirname(self.__filename) if self.__filename is not None else "."
        # select ANGEL input files whose name end with "*.out"
        fnames = QFileDialog.getOpenFileNames( # multiple files!
            self, "Angel2VTK - Choose input files", dir, "ANGEL input files (*.out)")
        # add each file to the input file list if it's not ALREADY there
        if not fnames.isEmpty():
            for fname in fnames:
                if self.inputFileList.findItems(fname,Qt.MatchExactly):
                    self.logTextEdit.appendHtml(
                        "%s <font color=red>already in the list, ignored</font>" % 
                        (fname))
                else:
                    self.inputFileList.addItem(fname)

    @pyqtSlot(bool)
    def on_actionRun_triggered(self):
        """Respond to the trrigered signal of the ``Angel2VTK_MainWindow.actionRun`` menu item.

        Just call the :py:meth:`fileConvert` method.
        """
        self.fileConvert()

    @pyqtSignature("")
    def on_convertPushButton_clicked(self):
        """Respond to the clicked signal of ``Angel2VTK_MainWindow.convertPushButton``.

        Just call the :py:meth:`fileConvert` method.
        """
        self.fileConvert()

    def fileConvert(self):
        """Convert ``*.out`` files to (``.root`` if in need and then to) VTK_ files in a thread.

        For real-time logging(?), the conversion is done in a separate thread.
        The contents of the ``Angel2VTK_MainWindow.inputFileList`` are cleared when done.
        """
        # input files as a python list
        inFilePyLST = [ str(self.inputFileList.item(idx).text())
                        for idx in range(self.inputFileList.count()) ]
        # print inFilePyLST
        if len(inFilePyLST) > 0:
            # print "calling processFiles()..."
            convTh = convertThread(inFilePyLST,
                                   digFolder=self.__digFolders,
                                   removeRoot=self.__removeRoots)
            # signal/slot for file conversion "running"
            convTh.convThRunning.connect(self.on_fileConvTh_running)
            # start the thread
            convTh.start()

    @pyqtSlot(bool)
    def on_fileConvTh_running(self, running):
        """Respond to the ``Angel2VTK_MainWindow.convThRunning`` signal and change some status flags.

        Toggle the status flag, change/restore the mouse pointer,
         and clear the ``Angel2VTK_MainWindow.inputFileList`` when finished (``running``=False).
        """
        if running:
            self.__running = True
            QApplication.setOverrideCursor(QCursor(Qt.WaitCursor))
        else:
            self.__running = False
            QApplication.restoreOverrideCursor()
            self.inputFileList.clear()

    @pyqtSignature("")
    def on_clearPushButton_clicked(self):
        """Respond to the clicked signal of ``Angel2VTK_MainWindow.clearPushButton``.

        Clear the ``Angle2VTK_MainWindow.inputFileList``.
        """
        self.inputFileList.clear()
        
    @pyqtSlot(bool)
    def on_actionClear_triggered(self):
        """Respond to the triggered signal of the ``Angel2VTK_MainWindow.actionClear`` menu item.

        Clear the inputFileList.
        """
        self.inputFileList.clear()

    ### Quit
    # NO NEED to def on_actionQuit_triggered(self);
    # it's connected to the closeEvent

    def closeEvent(self, event):
        """The method which is invoked when the close button is clicked.

        Save the status (position/size of the window and the checkboxes)
        using PyQt.QtCore.QSettings_ .

        .. _PyQt.QtCore.QSettings: http://pyqt.sourceforge.net/Docs/PyQt4/pyqt_qsettings.html
        """
        if self.__running:
                  self.statusBar().showMessage("Can't quit while running...",
                                               1000)
                  event.ignore()
        else:
            settings = QSettings()
            settings.setValue("MainWindow/Size", QVariant(self.size()))
            settings.setValue("MainWindow/Position", QVariant(self.pos()))
            settings.setValue("DigFolders", QVariant(self.__digFolders))
            settings.setValue("RemoveRoots", QVariant(self.__removeRoots))

    @pyqtSignature("int")
    def on_digFolderCheckBox_stateChanged(self, state):
        """Respond to the stateChanged signal of ``Angel2VTK_MainWindow.digFolderCheckBox``."""
        self.__digFolders = self.digFolderCheckBox.isChecked()
        # synchronize the corresponding menu item
        self.actionDigFolders.setChecked(self.digFolderCheckBox.isChecked())

    @pyqtSlot(bool)
    def on_actionDigFolders_toggled(self,boolVal):
        """Respond to the toggled signal of ``Angel2VTK_MainWindow.actionDigFolders`` menu item."""
        self.__digFolders = boolVal
        # synchronize the corresponding checkbox
        # NO tri-state, use QAbstractButton::setChecked() 
        self.digFolderCheckBox.setChecked(boolVal)
        #print "actionDigFolders toggled to ", boolVal

    @pyqtSignature("int")
    def on_removeRootCheckBox_stateChanged(self, state):
        """Respond to the stateChanged signal of ``Angel2VTK_MainWindow.removeRootCheckBox``."""
        self.__removeRoots = self.removeRootCheckBox.isChecked()
        # synchronize the corresponding menu item
        self.actionRemoveRoots.setChecked(self.removeRootCheckBox.isChecked())

    @pyqtSlot(bool)
    def on_actionRemoveRoots_toggled(self,boolVal):
        """Respond to the toggled signal of ``Angel2VTK_MainWindow.actionRemoveRoots`` menu item."""
        self.__removeRoots = boolVal
        # synchronize the corresponding checkbox
        # NO tri-state, use QAbstractButton::setChecked() 
        self.removeRootCheckBox.setChecked(boolVal)
        #print "actionRemoveRoots toggled to ", boolVal

    @pyqtSlot(bool)
    def on_actionAboutTheProgram_triggered(self):
        """Respond to the triggered signal of ``Angel2VTK_MainWindow.actionAboutTheProgram`` menu item.

        Just show the "About the program" dialog.
        """
        QMessageBox.about(self, "About the program",
            """<b>angel2vtk</b> v %s
            <p>Copyright &copy; 2014 JAEA? 
            All rights reserved.
            <p>This application can be used to convert
            Angel input files to VTK files.
            <p>Python %s - Qt %s - PyQt %s on %s""" % (
                __version__, platform.python_version(),
                QT_VERSION_STR, PYQT_VERSION_STR, platform.system()))

class LogStream(QObject):
    """A class for logging.

    This is a part of the logging method described in `an article on how-to-capture-output-of-pythons-interpreter-and-show-in-a-text-widget`_ .

    .. _`an article on how-to-capture-output-of-pythons-interpreter-and-show-in-a-text-widget`: \
    http://stackoverflow.com/questions/8356336/how-to-capture-output-of-pythons-interpreter-and-show-in-a-text-widget
    """

    #: An unbound signal with a string signature as the class attribute,
    #: created with :py:func:`pyqtSignal`.
    #: ("New signals should only be defined in sub-classes of :py:class:`QObject`.")
    textWritten = pyqtSignal(str)

    def write(self,text):
        """Emit a signal with the ``text`` passed."""
        self.textWritten.emit(str(text))

class logBuffer(QObject, IO.StringIO):
    """A class for logging w/ buffering.

    (keyword-)arguments are just passed to the initializers of the
    superclass.

    This is a part of the logging method described in
    http://stackoverflow.com/questions/14349563/how-to-get-non-blocking-real-time-behavior-from-python-logging-module-output-t
    """
    #: An unbound signal with a string signature as the class attribute,
    #: created with :py:func:`pyqtSignal`.
    #: ("New signals should only be defined in sub-classes of QObject.")
    bufferMessage = pyqtSignal(str)
    
    def __init__(self, *args, **kwargs):
        """Initialize the class just by calling superclass's initializers."""
        QObject.__init__(self)
        IO.StringIO.__init__(self, *args, **kwargs)
    
    def write(self, message):
        """Emit a signal with the text ``message`` passed."""
        if message:
            # emit a signal
            self.bufferMessage.emit(str(message.rstrip())) if Py3k else \
                self.bufferMessage.emit(unicode(message.rstrip()))
        IO.StringIO.write(self, message.rstrip())

class convertThread(threading.Thread, QObject):
    """A class for converting named ANGEL input files to VTK ones in a thread.

    This class was created to emit signals at the start and end of conversion
    (i.e. :py:meth:`run`) just for logging.

    For conversion, ROOT_ files are used as intermediate ones; these are created if needed,
    and are removed if told to do so.

    :Parameters:

    :py:data:`digFolder`
      a Boolean which specify whether to dig a folder and store the resultant files
      in it.

    :py:data:`removeRoot`
      a Boolean which specify whether to delte the intermediate ROOT_ file.
    """

    #: An unbound signal with a Boolean signature as the class attribute
    #: created with :py:func:`pyqtSignal`.
    #: ("New signals should only be defined in sub-classes of QObject.")
    convThRunning = pyqtSignal(bool)

    def __init__(self, inputFileList, digFolder=False, removeRoot=False):
        """Run the superclasss` initializers, start a logger, store the parameters as attributes.

        """
        # Individual initialization (as in the following lines) is
        # in need to bind the pyqtSignal to QObject.
        QObject.__init__(self)
        threading.Thread.__init__(self)
        self.logger = logging.getLogger(__name__)
        self.inputFileList = inputFileList
        self.digFolder=digFolder
        self.removeRoot=removeRoot

    def start(self):
        """Just start the superclass's :py:meth:`start` method."""
        super(convertThread, self).start()

    def run(self):
        """For each file, emit a starting signal, convert/store the data, then emit an ending signal."""
        self.convThRunning.emit(True)
        for inFileName in self.inputFileList:
            aFile = angelFile(inFileName)
            trackData = PHITS_TTrackData(aFile)
            (xnum,ynum,znum,enum,tnum) = trackData.getNumBins()
            (xtype,ytype,ztype,etype,ttype) = trackData.getTypes()
            # tnum = 1 and ttype = None if 't' does NOT exist.

            self.logger.info("input file : '%s'", inFileName)
            self.logger.info("  title = '%s'", trackData.getJobTitle())
            self.logger.info("  unit = '%s'",  trackData.getUnitInfo())
            self.logger.info("  axis = '%s'",  trackData.getAxisInfo())
            self.logger.info("  nbins(x,y,z,e,t) = (%d,%d,%d,%d,%d)",
                             xnum,ynum,znum,enum,tnum)
            self.logger.info("  types(x,y,z,e,t) = (%d,%d,%d,%d,%d)",
                             xtype,ytype,ztype,etype,ttype)
            if xtype > 2 or ytype > 2 or ztype > 2:
                self.logger.warning("[xyz]-type must be <= 2")
                sys.exit(2)
            elif xtype+ytype+ztype <= 4 and (xnum > 1 or ynum > 1 or znum > 1):
                self.logger.warning("implemented only for apparent"
                                    " IMAGE data at present,")
                sys.exit(2)

            # the number of digits in 't' bins
            #   it = 0,1,2, ..., 9 : #digits = 1
            #        0,1,2, .., 99 : #digits = 2
            #        0,1,2, ..,999 : #digits = 3
            ntdigits = int(math.ceil(math.log10(tnum)))
            VTKnamFmtStr = "%s%0" + "%d"%(ntdigits) + "d"
            self.logger.debug("VTKnamFmtStr = '%s'", VTKnamFmtStr)

            ##### output filenames
            # +--------------------------------+
            # |        outDirPlusBaseWOext     |
            # +----------------------+---------+
            # |   os.path.dirname()  |baseWOext|
            # v                      v         v
            # /some/where/in/the/path/something.out
            #
            # The root file is placed in the directory where .out file resides.
            rootFileName = os.path.splitext(inFileName)[0] + ".root"
            #
            if self.digFolder:
                dirName = os.path.splitext(inFileName)[0]
                baseWOext = os.path.basename(inFileName).replace(
                    os.path.splitext(inFileName)[1],'')
                outDirPlusBaseWOext = os.path.join(dirName, baseWOext)
                if not os.path.isdir(dirName):
                    self.logger.info("  digging the output directory '%s'", dirName)
                    os.mkdir(dirName)
            else:
                outDirPlusBaseWOext = os.path.splitext(inFileName)[0]
            self.logger.info("  root file : '%s'",   rootFileName)
            self.logger.info("  base name : '%s'\n", outDirPlusBaseWOext)
            # for the VTK group file name the extension is automatically added
            if ttype: # 't' exists
                outVTKG = VtkGroup( outDirPlusBaseWOext )

            # check whether the corresponding ROOT file exists or not
            if not os.path.exists(rootFileName):
                self.logger.info("  the corresponding root file DOES NOT exist."
                                 " creating...")
                cmdStr = "angel2root %s" % (inFileName)
                self.logger.debug("cmdStr = '%s'", cmdStr)
                os.system( cmdStr )

            # open (again?) the root file
            rootFile = ROOT.TFile( rootFileName, "read")

            #
            # reconstruct 3D data
            #
            (xmin,ymin,zmin) = trackData.getMinXYZ()
            if trackData.getAxisInfo() == "xy":  # axis="xy" -> (((t)z)xy)
                if ttype:
                    self.logger.info("  axis = \"xy\" -> pick 'z'(outer)"
                                     " then 't'(inner)")
                else:
                    self.logger.info("  axis = \"xy\" -> pick 'z'")
                xdel = trackData.getBinWidth('x')
                ydel = trackData.getBinWidth('y')
                zdel = trackData.getBinWidth('z')
                for it in range(tnum):
                    outVTKNam = VTKnamFmtStr % (outDirPlusBaseWOext,it) if ttype \
                                else outDirPlusBaseWOext
                    dataArray = np.zeros( (xnum,ynum,znum) )
                    for iz in range(znum):
                        hnum = 1 + iz*tnum + it
                        h2Name = "h%d" % (hnum)
                        self.logger.debug("it,iz,h = %d,%d,%s",it+1,iz+1,h2Name)
                        # get the TH2
                        hist = ROOT.gROOT.FindObject( h2Name )
                        for ix in range(xnum):
                            for iy in range(ynum):
                                bin = hist.GetBin(ix+1,iy+1) # 1-n in PHITS & ROOT, 0-(n-1) in Python
                                dataArray[ix,iy,iz] = hist.GetBinContent(bin)
                    # write a VTK file
                    imageToVTK(outVTKNam,
                               origin = (xmin,ymin,zmin),
                               spacing = (xdel,ydel,zdel),
                               cellData={"flux":dataArray} )
                    if ttype: # 't' exists
                        self.logger.info("  VTK image file '%s', t=[%g,%g](ns)",
                                     os.path.basename(outVTKNam),
                                     trackData.getLEdge('t',it),
                                         trackData.getUEdge('t',it))
                        # add it to the group file
                        # Note that PHITS in nsec and ...
                        outVTKG.addFile( filepath=outVTKNam+".vti",
                                         sim_time = trackData.getLEdge('t',it))
                    else:
                        self.logger.info("  VTK image file '%s'", outVTKNam)
                rootFile.Close()
                if ttype: # 't' exists
                    outVTKG.save()
            if self.removeRoot:
                self.logger.info("  removing intermediate ROOT file")
                os.remove(rootFileName)

        self.convThRunning.emit(False)


def GUImain():
    """The entry point for GUI.

    1. Command-line arguments are parsed using :py:mod:`python.argparse`, 
    2. a logger is created using :py:mod:`logging`,
    3. the form made of PyQt4_ is created and displayed, and then 
    4. the event loop is started.
    """
    # SETUPTOOLS automatic scripts expects a function w/o any arguments:
    # https://pythonhosted.org/setuptools/setuptools.html#automatic-script-creation
    # http://stackoverflow.com/questions/2853088/setuptools-not-passing-arguments-for-entry-points
    # So we rely on argparse here again...
    parser = argparse.ArgumentParser(
        description="Convert ANGEL input(.out) file(s) to VTK file(s)",
        epilog="With -d, the folder name will be the same as the basename of the input.",
        add_help=True)
    # no filename may be passed
    parser.add_argument("outFile", nargs='*', help="input ANGEL(.out) file")
    parser.add_argument("-d", "--dig-folder",
                        help="dig a folder in which files are stored",
                        action="store_true")
    parser.add_argument("-r", "--remove-root",
                        help="remove intermediate ROOT file",
                        action="store_true")
    args = parser.parse_args()

    app = QApplication(sys.argv)
    app.setOrganizationName("JAEA")
    app.setOrganizationDomain("jaea.go.jp")
    app.setApplicationName("angel2vtk")
    logger = logging.getLogger()
    logger.debug("angel.GUImain(): args.outFile = ", args.outFile)
    form = Angel2VTK_MainWindow(filelist=args.outFile,
                                digF=args.dig_folder,remR=args.remove_root)
    form.show()
    app.exec_()


def main():
    """The entry point for CUI.

    1. A logger is created, 
    2. command-line arguments are parsed using py:mod::argparse,
    3. the conversion thread is instantiated, and
    4. let the thead started.
    """
    # setup logging facility
    logger = logging.getLogger()                    # create logger
    logger.setLevel(logging.INFO)
    ch = logging.StreamHandler()                    # create console handler
    ch.setLevel(logging.INFO)
    logFormatter = logging.Formatter("%(message)s") # create formatter
    ch.setFormatter(logFormatter)                   # add formatter to ch
    logger.addHandler(ch)                           # add ch to logger

    # use argparse to parse the command-line arguments
    parser = argparse.ArgumentParser(
        description="Convert ANGEL(.out) file(s) to VTK file(s)",
        epilog="With -d, the folder name will be the same as the basename of the input.",
        add_help=True)
    parser.add_argument("outFile", nargs='+', help="input ANGEL(.out) file")
    parser.add_argument("-d", "--dig-folder",
                        help="dig a folder in which files are stored",
                        action="store_true")
    parser.add_argument("-r", "--remove-root",
                        help="remove intermediate ROOT file",
                        action="store_true")
    args = parser.parse_args()
    logging.debug("main(): args = %s", args)

    # create and run conversion thread
    convTh = convertThread(args.outFile, 
                           digFolder=args.dig_folder,
                           removeRoot=args.remove_root)
    convTh.start()

if __name__ == "__main__":
    main()
